# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/gui/__init__.py
""" Soar GUI Modules.

Contains the modules for building and displaying Soar's GUI, as well as all GIFs used for icons or animations, and
widgets for plot generation, etc.
"""
from soar import __version__