# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/sim/__init__.py
""" Simulation modules, containing general 2D geometry methods, graphical world objects with collision detection, etc.
"""
