# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/robot/__init__.py
""" Soar robot modules, containing the BaseRobot class, example robot subclasses, and bundled implementations like
PioneerRobot.
"""
